

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#include <string.h>

#define _USE_MATH_DEFINES 1
#include <math.h>

//for setting array sizes
#define MAX_TREES 50000
#define MAX_X_CELLS 100
#define MAX_Y_CELLS 100

//cell dimension, in m.
#define CELL_LENGTH 8

//number of harvest trials to do before giving up.
#define MAX_ATTEMPTS 100
//relative difference between target prop cut and actual prop cut to accept.
#define ACCEPTABLE_CLOSENESS 0.01

#define INPUTFILENAME "gap_harvest_input.txt"
#define CUTOUTPUTFILENAME "gap_harvest_cut_trees.txt"
#define RETAINEDOUTPUTFILENAME "gap_harvest_retained_trees.txt"
#define SUMMARYOUTPUTFILENAME "gap_harvest_summary.txt"

typedef struct {
	char input[300];
	char cutoutput[300];
	char retainedoutput[300];
	char summaryoutput[300];
} files;
files f;

typedef struct {
  float plot_x;
  float plot_y;
  float prop_ba_to_cut;
  float lognorm_mu;
  float lognorm_sigma;
  char filepath[200];
} arg;

arg args;

typedef struct {
  float x;
  float y;
  int x_cell;
  int y_cell;
  int species;
  int type;
  float dbh;
  float ba;
  float height;
  int sort_order;
  int prev_in_cell;
  int next_in_cell;
  int prev_in_pop;
  int next_in_pop;
} tree;

tree unallocated[MAX_TREES];
tree * cut[MAX_TREES];
tree * retained[MAX_TREES];

typedef struct {
  int first_in_cell;
  int n_trees;
} cells;

typedef struct {
  cells cell[MAX_X_CELLS][MAX_Y_CELLS];
  int x_cells;
  int y_cells;
  int n_cut;
  int n_retained;
  int n_left;
  int n_patches;
  int n_attempts;
  float ba_cut;
  float ba_retained;
  float ba_left;
  float target_ba_cut;
  float target_ba_retained;
  float closeness_to_target;
  float sum_patch_area;
  float sumsq_patch_area;
} pop;

pop grid;

typedef struct {
  float x;
  float d;
  float p;
} float_set;

char FirstLine[100];
char SecondLine[100];

float_set cdf[10000];

void make_cdf_table();
void initialization();
int read_file();
void set_criteria();
void cut_or_retain_patch(tree **list, int * n, float * ba);
double inverse_normal_fn(double p);
double normal_cdf(const double x);
void make_links(int index);
void cut_or_retain_remaining(tree **list, int * n, float * ba);
void write_files();
void write_empty_cut_trees_file();

/**************************************************************************/
int main( int argc, char * argv[] )
{

  int curr_timestep;


  //Mandatory arguments passed are as follows:
  //(1) plot x length (m)
  //(2) plot y length (m)
  //(3) proportion of BA to cut
  //(4) mu parameter for lognormal distribution of patch area
  //(5) sigma parameter for lognormal distribution of patch area
  //(6) path to input and output files

  //read argument values
  if (argc != 7) {
    printf("Wrong number of arguments passed! (Should be 6.)\n");
    exit(99);
  }
  //assign to variables
  args.plot_x = (float) strtod (argv[1], NULL);
  args.plot_y = (float) strtod (argv[2], NULL);
  args.prop_ba_to_cut = (float) strtod (argv[3], NULL);
  args.lognorm_mu = (float) strtod (argv[4], NULL);
  args.lognorm_sigma = (float) strtod (argv[5], NULL);
  strcpy(args.filepath, argv[6]);

  sprintf(f.input,"%s%s",args.filepath, INPUTFILENAME);
  sprintf(f.cutoutput,"%s%s",args.filepath, CUTOUTPUTFILENAME);
  sprintf(f.retainedoutput,"%s%s",args.filepath, RETAINEDOUTPUTFILENAME);
  sprintf(f.summaryoutput,"%s%s",args.filepath, SUMMARYOUTPUTFILENAME);


  //set some initial variables
  grid.closeness_to_target = 99;
  grid.n_attempts = 0;

  //do the harvesting, repeating if necessary until acceptability criteria is met
  while (grid.closeness_to_target > ACCEPTABLE_CLOSENESS) {

    //set up initial values
    initialization();

    //read tree input data
    curr_timestep = read_file();

    if (curr_timestep != 1) {
     //if this isn't the first timstep, write a null list of cut trees and exit
      write_empty_cut_trees_file();
      exit(0);
    }

    //set the target cut and retention BAs
    set_criteria();

    //alternately cut and retain patches of random size until one of targets is met
    while ((grid.ba_cut < grid.target_ba_cut) && (grid.ba_retained < grid.target_ba_retained) && (grid.n_left > 0)) {

      cut_or_retain_patch(cut, &(grid.n_cut), &(grid.ba_cut));
      if (grid.n_left > 0)
        cut_or_retain_patch(retained, &(grid.n_retained), &(grid.ba_retained));

      grid.n_patches ++;

    } //end while

    if (grid.ba_cut > grid.target_ba_cut) {
      //cut target met, retain remainder
      cut_or_retain_remaining(retained, &(grid.n_retained), &(grid.ba_retained));
    }
    else {
      //retention target met, cut remainder
      cut_or_retain_remaining(cut, &(grid.n_cut), &(grid.ba_cut));
    }

    //check how close we got to our original target
    grid.closeness_to_target = (fabs(grid.ba_cut - grid.target_ba_cut)/grid.target_ba_cut);

  } //end while

  //write output treelists and summary data
  write_files();

  return 0;
}

/**************************************************************************/
void initialization() {

  int x_loop, y_loop, tree_loop;

  if (grid.n_attempts == 0) {

    //random seed
    srand(time(NULL));

    //grid size
    grid.x_cells = args.plot_x / ((int) CELL_LENGTH) + 1;
    grid.y_cells = args.plot_y / ((int) CELL_LENGTH) + 1;

    if ((grid.x_cells > MAX_X_CELLS) || (grid.y_cells > MAX_Y_CELLS)) {
      printf("Grid is too large.\n");
      exit(99);
    } //end if

  } //end if (first attempt)

  for (x_loop=0;x_loop<grid.x_cells;x_loop++)
    for (y_loop=0;y_loop<grid.y_cells;y_loop++) {
      grid.cell[x_loop][y_loop].first_in_cell = -1;
      grid.cell[x_loop][y_loop].n_trees = 0;
    } //end for

  for (tree_loop=0; tree_loop<MAX_TREES; tree_loop++) {

    cut[tree_loop] = 0;
    retained[tree_loop] = 0;

    unallocated[tree_loop].x = 0;
    unallocated[tree_loop].y = 0;
    unallocated[tree_loop].species = 0;
    unallocated[tree_loop].type = 0;
    unallocated[tree_loop].dbh = 0;
    unallocated[tree_loop].ba = 0;
    unallocated[tree_loop].height = 0;
    unallocated[tree_loop].x_cell = 0;
    unallocated[tree_loop].y_cell = 0;
    unallocated[tree_loop].sort_order = 0;
    unallocated[tree_loop].prev_in_cell = 0;
    unallocated[tree_loop].next_in_cell = 0;
    unallocated[tree_loop].prev_in_pop = 0;
    unallocated[tree_loop].next_in_pop = 0;

  } //end for

  grid.n_patches = 0;
  grid.sum_patch_area = 0;
  grid.sumsq_patch_area = 0;
  grid.n_cut = 0;
  grid.n_retained = 0;
  grid.n_left = 0;
  grid.ba_cut = 0;
  grid.ba_retained = 0;
  grid.ba_left = 0;

  if (grid.n_attempts > MAX_ATTEMPTS) {
    printf("Gave up after too many unsuccessful attempts to approach target.\n");
    exit(99);
  }

  grid.n_attempts ++;

  return;
}

/**************************************************************************/
int read_file() {
  //returns the current timestep number from the first line of the input file


  FILE *ifp;

  int timestep;

  float X,Y,Diam,Height;
  int Species, Type;
  char line[100];

  int x_cell, y_cell;

  int temp_index;

  ifp = fopen(f.input, "r");
    if (ifp==NULL) {
      fprintf(stderr,"Can't open input file\n");
      exit(99);
    }

//scan input data

  //first line is timestep info
  fgets(FirstLine, 100, ifp);
  sscanf(FirstLine, "%d\t%*d\n", &timestep);
  //second line is the column headings
  fgets(SecondLine, 100, ifp);

  //subsequent lines are the trees
  grid.n_left=0;

  while (fgets(line, 100, ifp)) {

    sscanf(line,"%f\t%f\t%d\t%d\t%f\t%f\n",
        &X,
        &Y,
        &Species,
        &Type,
        &Diam,
        &Height);

    if (Type >= 2) { //saplings, adults, and snags
      grid.n_left++; //n_left is number of unallocated trees

      //assign data to structure
      unallocated[grid.n_left].x = X;
      unallocated[grid.n_left].y = Y;
      unallocated[grid.n_left].species = Species;
      unallocated[grid.n_left].type = Type;
      unallocated[grid.n_left].dbh = Diam;
      unallocated[grid.n_left].ba = pow(Diam/200.0,2);
      unallocated[grid.n_left].height = Height;

      unallocated[grid.n_left].sort_order=grid.n_left; //for sorting back to original order, but isn't used now

      x_cell = ((int) X) / ((int) CELL_LENGTH);
      y_cell = ((int) Y) / ((int) CELL_LENGTH);

      if ((x_cell >= grid.x_cells) || (y_cell >= grid.y_cells)) {
        printf("Tree is outside grid boundary.\n");
        printf("(%4.4f, %4.4f)\n", X, Y);
        exit(99);
      } //end if

      unallocated[grid.n_left].x_cell = x_cell;
      unallocated[grid.n_left].y_cell = y_cell;

      //set the linked list indices for the grid cell the tree is in
      if (grid.cell[x_cell][y_cell].first_in_cell == -1) { //if this is first tree in cell
        grid.cell[x_cell][y_cell].first_in_cell = grid.n_left;
        unallocated[grid.n_left].prev_in_cell = -1;
      } //end if
      else { // not first tree in cell
        temp_index = grid.cell[x_cell][y_cell].first_in_cell;
        while (unallocated[temp_index].next_in_cell != -1) {
         temp_index = unallocated[temp_index].next_in_cell;
        } //end while

      unallocated[temp_index].next_in_cell = grid.n_left;
      unallocated[grid.n_left].prev_in_cell = temp_index;

      } //end else

      unallocated[grid.n_left].next_in_cell = -1;
      grid.cell[x_cell][y_cell].n_trees ++;

      //set the linked list indices for the population (it's the same order they're read in)
      unallocated[grid.n_left].next_in_pop = -1;
      unallocated[grid.n_left].prev_in_pop = grid.n_left - 1;
      unallocated[grid.n_left-1].next_in_pop = grid.n_left;

      grid.ba_left += unallocated[grid.n_left].ba;

      if (grid.n_left == MAX_TREES) {
        fprintf(stderr, "Number of input trees exceeds max length of array!\n");
        exit(99);
      } //end if

    } //end if (adult tree)

  } //end while

  fclose(ifp);

  return timestep;
}

/**************************************************************************/
void set_criteria() {

  //convert from cut proportion to actual ba (in m^2, but not per ha)
  grid.target_ba_cut = args.prop_ba_to_cut * grid.ba_left;
  grid.target_ba_retained = (1.0 - args.prop_ba_to_cut) * grid.ba_left;

 return;
}

/**************************************************************************/
void cut_or_retain_patch(tree **list, int * n, float * ba) {
  //this function will assign trees to either the cut or retained list
  //(the actual list is specified by the pointers passed)

  int i, centre_index, rand_count, curr_index;
  int min_x_cell, max_x_cell, min_y_cell, max_y_cell, curr_x_cell, curr_y_cell;
  float x_dist, y_dist;

  float patch_area, patch_radius;

  //get random patch radius

  //log-normal patch-area distribution
  patch_area = exp(args.lognorm_mu + args.lognorm_sigma
    * inverse_normal_fn((1.0*(rand()+1))/(1.0*(RAND_MAX+1))));

  //for calculating mean and sd of patch area in summary stats
  grid.sum_patch_area += patch_area;
  grid.sumsq_patch_area += patch_area * patch_area;

  patch_radius = pow(patch_area/M_PI, 0.5);

  if (patch_radius < 0) {
    patch_radius = 0.0;
  } //end if
  else if ((patch_radius > args.plot_x/4.0) || (patch_radius > args.plot_y/4.0)) {
    //max radius for patch is 1/4 of smaller plot dimension
    patch_radius = args.plot_x/4.0;
    if (patch_radius > args.plot_y/4.0) {
      patch_radius = args.plot_y/4.0;
    }
  } //end else

  //get random tree at centre
  centre_index=0;
  rand_count = 1 + rand() % grid.n_left; //random # b/w 1 and n_left

  for (i=0; i<rand_count; i++) {
   centre_index = unallocated[centre_index].next_in_pop;
  }

  //remove centre tree from unallocated list, and add to cut/retain list
  list[*n] = &(unallocated[centre_index]);
  (*n) ++;
  (*ba) += unallocated[centre_index].ba;

  grid.cell[unallocated[centre_index].x_cell][unallocated[centre_index].x_cell].n_trees --;
  grid.n_left --;
  grid.ba_left -= unallocated[centre_index].ba;

  //re-connect the linked list
  make_links(centre_index);

  if (patch_radius > 1E-6) { //if non-zero radius, then cut trees around centre tree

    //find upper and lower bounds of grid cells around centre tree for searching
    min_x_cell = ((int)(args.plot_x + unallocated[centre_index].x - patch_radius - 1)
      / (int) CELL_LENGTH) % grid.x_cells;
    max_x_cell = ((int)(unallocated[centre_index].x + patch_radius + 1)
      / (int) CELL_LENGTH) % grid.x_cells;
    min_y_cell = ((int)(args.plot_y + unallocated[centre_index].y - patch_radius - 1)
      / (int) CELL_LENGTH) % grid.y_cells;
    max_y_cell = ((int)(unallocated[centre_index].y + patch_radius + 1)
      / (int) CELL_LENGTH) % grid.y_cells;

    //move through the cells, checking each tree to see whether it's included
    curr_x_cell = min_x_cell;
    while (curr_x_cell != ((max_x_cell+1)%grid.x_cells)) {
     curr_y_cell = min_y_cell;
     while (curr_y_cell != ((max_y_cell+1)%grid.y_cells)) {

       curr_index = grid.cell[curr_x_cell][curr_y_cell].first_in_cell;

       while (curr_index != -1) {

         //calculate distance and act if it's within range
         x_dist = fabs(unallocated[centre_index].x - unallocated[curr_index].x);
         if (args.plot_x - x_dist < x_dist)
           x_dist = args.plot_x - x_dist;

         y_dist = fabs(unallocated[centre_index].y - unallocated[curr_index].y);
         if (args.plot_y - y_dist < y_dist)
           y_dist = args.plot_y - y_dist;

         if ((pow(x_dist*x_dist + y_dist*y_dist, (float)0.5) < patch_radius) && (centre_index != curr_index)) {
            //tree is within radius (and not the centre tree);
            //remove it from unallocated list and add to appropriate cut/retain list
            list[*n] = &(unallocated[curr_index]);
            (*n) ++;
            (*ba) += unallocated[curr_index].ba;

            grid.cell[unallocated[curr_index].x_cell][unallocated[curr_index].x_cell].n_trees --;
            grid.n_left --;
            grid.ba_left -= unallocated[curr_index].ba;

            make_links(curr_index);

           } //end if

        curr_index = unallocated[curr_index].next_in_cell;
       } //end while another tree in cell

       curr_y_cell = (curr_y_cell+1) % grid.y_cells;
     } //end y_cell while
     curr_x_cell = (curr_x_cell+1) % grid.x_cells;
    } //end x_cell while

  } //end if radius > 0

  return;
}

/*****************************************************************************/
double inverse_normal_fn(double p) {
  //parameter passed is random P value (0 to 1)
  //function returns corresponding value from inverse cumulative normal distribution

  /* Coefficients in rational approximations. */
  static const double a[] =
  {
          -3.969683028665376e+01,
           2.209460984245205e+02,
          -2.759285104469687e+02,
           1.383577518672690e+02,
          -3.066479806614716e+01,
           2.506628277459239e+00
  };
  static const double b[] =
  {
          -5.447609879822406e+01,
           1.615858368580409e+02,
          -1.556989798598866e+02,
           6.680131188771972e+01,
          -1.328068155288572e+01
  };
  static const double c[] =
  {
          -7.784894002430293e-03,
          -3.223964580411365e-01,
          -2.400758277161838e+00,
          -2.549732539343734e+00,
           4.374664141464968e+00,
           2.938163982698783e+00
  };
  static const double d[] =
  {
          7.784695709041462e-03,
          3.224671290700398e-01,
          2.445134137142996e+00,
          3.754408661907416e+00
  };

  double q, r;

  if (p <= 0 || p >= 1) {
    return 0.0;
  }
  else if (p < 0.02425) {
    /* Rational approximation for lower region */
    q = sqrt(-2*log(p));
    return (((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5]) /
            ((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1);
  }
  else if (p > 0.97575) {
    /* Rational approximation for upper region */
    q  = sqrt(-2*log(1-p));
    return -(((((c[0]*q+c[1])*q+c[2])*q+c[3])*q+c[4])*q+c[5]) /
            ((((d[0]*q+d[1])*q+d[2])*q+d[3])*q+1);
  }
  else {
    /* Rational approximation for central region */
    q = p - 0.5;
    r = q*q;
    return (((((a[0]*r+a[1])*r+a[2])*r+a[3])*r+a[4])*r+a[5])*q /
            (((((b[0]*r+b[1])*r+b[2])*r+b[3])*r+b[4])*r+1);
  }

}
/*****************************************************************************/

double normal_cdf(const double x) {
  //returns value from cumulative normal distribution.

  const double b1 =  0.319381530;
  const double b2 = -0.356563782;
  const double b3 =  1.781477937;
  const double b4 = -1.821255978;
  const double b5 =  1.330274429;
  const double p  =  0.2316419;
  const double c  =  0.39894228;

  if(x >= 0.0) {
      double t = 1.0 / ( 1.0 + p * x );
      return (1.0 - c * exp( -x * x / 2.0 ) * t *
      ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
  }
  else {
      double t = 1.0 / ( 1.0 - p * x );
      return ( c * exp( -x * x / 2.0 ) * t *
      ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
    }
}

/*****************************************************************************/
void make_links(int index) {
  //removes tree [index] from the linked list of unallocated trees

  if (grid.cell[unallocated[index].x_cell][unallocated[index].y_cell].first_in_cell == index) {
    grid.cell[unallocated[index].x_cell][unallocated[index].y_cell].first_in_cell =
      unallocated[index].next_in_cell;
  } //end if first in cell
  else {
    unallocated[unallocated[index].prev_in_cell].next_in_cell =
      unallocated[index].next_in_cell;
  } //end else
  if (unallocated[index].next_in_cell != -1) {
   unallocated[unallocated[index].next_in_cell].prev_in_cell =
    unallocated[index].prev_in_cell;
  }

  unallocated[unallocated[index].prev_in_pop].next_in_pop =
      unallocated[index].next_in_pop;
  if (unallocated[index].next_in_pop != -1) {
   unallocated[unallocated[index].next_in_pop].prev_in_pop =
    unallocated[index].prev_in_pop;
  }

  return;
}

/**************************************************************************/
void cut_or_retain_remaining(tree **list, int * n, float * ba) {
  //goes through the list of unallocated trees and copies everything to either
  //cut or retained list (depending on pointer passed).

  int index;

  index=0;

  while (grid.n_left > 0) {
   index = unallocated[index].next_in_pop;

   //remove tree from unallocated list,
   //and add to appropriate list
   list[*n] = &(unallocated[index]);
   (*n) ++;
   (*ba) += unallocated[index].ba;

   grid.cell[unallocated[index].x_cell][unallocated[index].x_cell].n_trees --;
   grid.n_left --;
   grid.ba_left -= unallocated[index].ba;

   make_links(index);

  } //end while

  return;
}

/*****************************************************************************/
void write_files() {
  //writes lists of cut and retained trees, as well as a file with summary statistics

  int i;
  FILE *cut_ofp, *ret_ofp, *summary_ofp;

  cut_ofp = fopen(f.cutoutput,"w");
  if (cut_ofp==NULL) {
    printf("Can't open cut trees file.\n");
    exit(99);
  }

  //write header lines
  fputs(FirstLine, cut_ofp);
  fputs(SecondLine, cut_ofp);

  for(i=0;i<grid.n_cut;i++) {

      fprintf(cut_ofp,"%4.4f\t%4.4f\t%d\t%d\t%4.4f\t%4.4f\n",
        (cut[i]) -> x,
        (cut[i]) -> y,
        (cut[i]) -> species,
        (cut[i]) -> type,
        (cut[i]) -> dbh,
        (cut[i]) -> height);

    } //end for

  fclose(cut_ofp);

  ret_ofp = fopen(f.retainedoutput,"w");
  if (ret_ofp==NULL) {
    printf("Can't open retained trees file.\n");
    exit(99);
  }

  //write header lines
  fputs(FirstLine, ret_ofp);
  fputs(SecondLine, ret_ofp);

  for(i=0;i<grid.n_retained;i++) {

      fprintf(ret_ofp,"%4.4f\t%4.4f\t%d\t%d\t%4.4f\t%4.4f\n",
        (retained[i]) -> x,
        (retained[i]) -> y,
        (retained[i]) -> species,
        (retained[i]) -> type,
        (retained[i]) -> dbh,
        (retained[i]) -> height);

    } //end for

  fclose(ret_ofp);


  summary_ofp = fopen(f.summaryoutput,"w");
  if (summary_ofp==NULL) {
    printf("Can't open summary output file.\n");
    exit(99);
  }

  fprintf(summary_ofp,"n_cut\t%d\nba_cut\t%4.4f\nn_retained\t%d\nba_retained\t%4.4f\nn_attempts\t%d\nn_patches\t%d\nmean_patch_area\t%4.4f\nsd_patch_area\t%4.4f\nprop_ba_cut\t%4.4f\ntarget_prop_ba_cut\t%4.4f\n",
    grid.n_cut,
    grid.ba_cut / (args.plot_x*args.plot_y / 10000.0),
    grid.n_retained,
    grid.ba_retained / (args.plot_x*args.plot_y / 10000.0),
    grid.n_attempts,
    grid.n_patches,
    grid.sum_patch_area / (2.0*grid.n_patches), //mean patch area
    pow((grid.sumsq_patch_area - pow(grid.sum_patch_area,2)/(2.0*grid.n_patches))/(2.0*grid.n_patches - 1.0), 0.5), //sd patch area
    grid.ba_cut / (grid.ba_cut + grid.ba_retained),
    args.prop_ba_to_cut
    );

  fclose(summary_ofp);

  return;

}

/*****************************************************************************/
void write_empty_cut_trees_file() {
  //writes null list of cut trees

  FILE *cut_ofp;

  cut_ofp = fopen(f.cutoutput,"w");
  if (cut_ofp==NULL) {
    printf("Can't open cut trees file.\n");
    exit(99);
  }

  //write header lines
  fputs(FirstLine, cut_ofp);
  fputs(SecondLine, cut_ofp);

  fclose(cut_ofp);

  return;

}
